/* PersistentDS.p  
    Last Updated- DL 5/05/04 
    
    Overview of code:
    This demonstration 4GL code is designed to show passing a static ProDataSet and 
    the use of Open Client Procedure Objects accessing persistently running 4GL code.  
    The code can be run on AppServer and then called by .NET clients using the OpenEdge 
    .NET Open Client.  
    
    From an implementation perspective this design that reduces the amount of reads to the physical
    database by keeping the records in a ProDataSet persistently would only make sense if there
    is a high degree of complexity to the data building up the data cache. 
    
    Description of the 4GL procedures and user defined functions:
    This sample code is used for demonstration purposes and is designed to be run persistent.
    Access to the database is minimized by reading a large number records into a master data set.
    Requests can then be made to return a subset of records that are available in a master data set.
    
    The master data set is filled by calling procedure fillMaster.  Records are retrived 
    from the Sports2000 database.  The data set contains orders 
    and associated orderline records.  A high/low range is input to the procedure specifying the
    order records to be selected.  The procedure can be recalled to refresh the records and to 
    retrive a different range of order records from the database.
    
    The subset data set retrives records from the master data set by calling procedure
    returnSubset.  This minimizes the number of calls needed to the database.  
    A high/low range of orders is be specified.  However, if the requested orders 
    are outside of the range of records in the master data set then the records
    are not found, even if they exist in the database. 
    
    The DATETIME-TZ variables, along with the setLastUpdate and getLastUpdate 
    functions are used to track when the records in the master data set were retrived.  The
    Procedure returnSubset contains code to automatically call procedure 
    fillMaster after a certain period of time has elapsed since the last update.
    For demonstration purposes the elapsed time is 5 min.
*/
 
/* Master data definitions */ 
{OrderTables.i}
{OrderDS.i}

/* Subset data definitions */ 
DEFINE TEMP-TABLE ttSubsetOrder LIKE ttOrder.
DEFINE TEMP-TABLE ttSubsetOLine LIKE ttOLine
   BEFORE-TABLE ttSubsetOlineBefore.
DEFINE DATASET dsSubsetOrder FOR ttSubsetOrder, ttSubsetOLine
    DATA-RELATION OrderLine FOR ttSubsetOrder, ttSubsetOLine
      RELATION-FIELDS (OrderNum, OrderNum).

DEFINE VARIABLE dtzLastUpdate AS DATETIME-TZ NO-UNDO INITIAL ?.
DEFINE VARIABLE iMasterGE AS INTEGER  NO-UNDO INITIAL 0.
DEFINE VARIABLE iMasterLE AS INTEGER  NO-UNDO INITIAL 0.

/* refresh every 5 minutes - in milliseconds */ 
DEFINE VARIABLE iRefreshMill AS INTEGER NO-UNDO INITIAL 30000.

FUNCTION getLastUpdate RETURNS DATETIME-TZ PRIVATE () FORWARD.
FUNCTION setLastUpdate RETURNS DATETIME-TZ PRIVATE () FORWARD.

PROCEDURE fillMaster:
/*------------------------------------------------------------------------------
  Purpose: Populate the master 
  Parameters: 
  Notes:       
------------------------------------------------------------------------------*/
    DEFINE INPUT  PARAMETER piGE AS INTEGER NO-UNDO.
    DEFINE INPUT  PARAMETER piLE AS INTEGER NO-UNDO.
    
    DEFINE VARIABLE hdsOrder AS HANDLE     NO-UNDO.
    DEFINE VARIABLE iBuff     AS INTEGER    NO-UNDO.

    DEFINE QUERY qOrder FOR Order.

    DEFINE DATA-SOURCE srcOrder FOR QUERY qOrder Order KEYS (OrderNum).
    DEFINE DATA-SOURCE srcOline FOR OrderLine.

    OUTPUT TO "C:\OpenEdge\WRK\fillMasterOutputC1930.txt".
    DISPLAY "PersistentDS.p - fillMaster" SKIP.
    hdsOrder = DATASET dsOrder:HANDLE.
    hdsOrder:EMPTY-DATASET.

    /* Prepare the query */
    QUERY qOrder:QUERY-PREPARE("FOR EACH Order NO-LOCK WHERE " +
                               "Order.OrderNum GE " + STRING(piGE) + " AND " + 
                               "Order.OrderNum LE " + STRING(piLE)).
    /* QUERY qOrder:QUERY-PREPARE("FOR EACH Order WHERE Order.OrderNum GE 1 AND Order.OrderNum LE 5"). */

    /* Attach datasources before fill */
    BUFFER ttOrder:ATTACH-DATA-SOURCE(DATA-SOURCE srcOrder:HANDLE).
    BUFFER ttOLine:ATTACH-DATA-SOURCE(DATA-SOURCE srcOline:HANDLE).

    hDSOrder:FILL().
    ASSIGN
        iMasterGE = piGE
        iMasterLE = piLE.
    setLastUpdate().

    OUTPUT CLOSE. 
    RETURN.
END PROCEDURE.

PROCEDURE returnSubset:
/*------------------------------------------------------------------------------
  Purpose: Return a subset of the records in the master  
  Parameters: 
  Notes:       
------------------------------------------------------------------------------*/
    DEFINE INPUT  PARAMETER piSubGE  AS INTEGER    NO-UNDO.
    DEFINE INPUT  PARAMETER piSubLE  AS INTEGER    NO-UNDO.
    DEFINE OUTPUT PARAMETER DATASET FOR dsSubsetOrder.

    DEFINE VARIABLE dtzCurMasterUpdate AS DATETIME-TZ NO-UNDO.
    DEFINE VARIABLE dtzRefreshtime AS DATETIME-TZ NO-UNDO.
    DEFINE VARIABLE hdsSubsetOrder AS HANDLE     NO-UNDO.
    DEFINE VARIABLE iBuff     AS INTEGER    NO-UNDO.

    DEFINE QUERY qSubsetOrder FOR ttOrder.

    DEFINE DATA-SOURCE srcOrder FOR QUERY qSubsetOrder ttOrder KEYS (OrderNum).
    DEFINE DATA-SOURCE srcOline FOR ttOLine.

    OUTPUT TO "C:\OpenEdge\WRK\returnSubsetOutputC1930.txt".
    DISPLAY "PersistentDS.p - returnSubset" SKIP.

    IF (dtzLastUpdate NE ?) THEN DO: 

        hdsSubsetOrder = DATASET dsSubsetOrder:HANDLE.
        hdsSubsetOrder:EMPTY-DATASET.
    
        /* Check to see if master should be refreshed ... */
        dtzCurMasterUpdate = getLastUpdate().

        /* Check if current time greateer than (saved time + refresh factor) */ 
        dtzRefreshtime = DATETIME-TZ( DATE(dtzCurMasterUpdate), 
                                      MTIME(dtzCurMasterUpdate) + iRefreshMill, 
                                      TIMEZONE(dtzCurMasterUpdate)). 
        IF ( DATETIME-TZ(TODAY, MTIME, TIMEZONE) GT dtzRefreshtime ) THEN 
            /* Update master dataset */  
            RUN fillMaster(iMasterGE, iMasterLE).
    
        /* Prepare the query */
        QUERY qSubsetOrder:QUERY-PREPARE("FOR EACH ttOrder WHERE " +
                                   "ttOrder.OrderNum GE " + STRING(piSubGE) + " AND " + 
                                   "ttOrder.OrderNum LE " + STRING(piSubLE)).
    
        /* Attach datasources before fill */
        BUFFER ttSubsetOrder:ATTACH-DATA-SOURCE(DATA-SOURCE srcOrder:HANDLE).
        BUFFER ttSubsetOLine:ATTACH-DATA-SOURCE(DATA-SOURCE srcOline:HANDLE).
    
        hdsSubsetOrder:FILL().
    
        /*Detach datasources after fill */
        DO iBuff = 1 TO DATASET dsSubsetOrder:NUM-BUFFERS:
             DATASET dsSubsetOrder:GET-BUFFER-HANDLE(iBuff):DETACH-DATA-SOURCE().
        END.
        
    END.

    OUTPUT CLOSE. 
    RETURN.
END PROCEDURE.

FUNCTION getLastUpdate RETURNS DATETIME-TZ PRIVATE ():
/*------------------------------------------------------------------------------
  Purpose: Set the last date/time the master was updated from the database 
------------------------------------------------------------------------------*/
    RETURN dtzLastUpdate.
END FUNCTION.

FUNCTION setLastUpdate RETURNS DATETIME-TZ PRIVATE ():
/*------------------------------------------------------------------------------
  Purpose: Set the last date/time the master was updated from the database 
------------------------------------------------------------------------------*/
    dtzLastUpdate = DATETIME-TZ(TODAY, MTIME, TIMEZONE). 
    RETURN dtzLastUpdate.
END FUNCTION.
    
